# Licensed under the GPLv3 - see LICENSE
from astropy import units as u

from ..base.file_info import (info_item, InfoBase,
                              FileReaderInfo, StreamReaderInfo)


def file_size(fh):
    offset = fh.tell()
    try:
        return fh.seek(0, 2)
    finally:
        fh.seek(offset)


class GSBTimeStampInfo(InfoBase):
    """Standardized information on a timestamp file reader.

    The ``info`` descriptor has a number of standard attributes, which are
    determined from arguments passed in opening the file, from the first header
    (``info.header0``) and from possibly scanning the file to determine the
    duration of frames.
    """
    attr_names = ('format', 'mode', 'number_of_frames', 'frame_rate',
                  'start_time', 'readable', 'missing', 'errors', 'warnings')
    mode = info_item(needs='header0', doc=(
        "Mode in which data was taken: 'phased' or 'rawdump'."))
    # We do not subclass FileReaderInfo, since the logic is a bit different.
    # Still, these ones can be used directly.
    start_time = FileReaderInfo.start_time
    frame_rate = FileReaderInfo.frame_rate
    checks = FileReaderInfo.checks
    errors = FileReaderInfo.errors
    warnings = FileReaderInfo.warnings

    # We cannot know whether the stream is readable without the raw files.
    readable = None
    missing = info_item(default={
        'raw': 'need raw binary files for the stream reader'}, copy=True)

    @info_item
    def header0(self):
        with self._parent.temporary_offset(0) as fh:
            return fh.read_timestamp()

    @info_item(needs='header0')
    def format(self):
        return 'gsb'

    @info_item(needs='header0')
    def number_of_frames(self):
        with self._parent.temporary_offset() as fh:
            fh_size = fh.seek(0, 2)
            # Guess based on a fixed header size.  In reality, this
            # may be an overestimate as the headers can grow in size,
            # or an underestimate as the last header may be partial.
            # So, search around to be sure.
            guess = max(fh_size // self.header0.nbytes, 1)
            while self.header0.seek_offset(guess) > fh_size:
                guess -= 1
            while self.header0.seek_offset(guess) < fh_size:
                guess += 1

            # Now see if there is indeed a nice header before.
            fh.seek(self.header0.seek_offset(guess-1))
            line_tuple = fh.readline().split()
            # But realize that sometimes an incomplete header is written.
            if (len(" ".join(line_tuple))
                    < len(" ".join(self.header0.words))):
                self.warnings['number_of_frames'] = (
                    'last header is incomplete and is ignored')
                retry = True
            else:
                # Check last header is readable.
                try:
                    self.header0.__class__(line_tuple).time
                except Exception as exc:
                    self.warnings['number_of_frames'] = (
                        'last header failed to read ({}) and is ignored'
                        .format(str(exc)))
                    retry = True
                else:
                    retry = False
            if retry:
                guess -= 1
                fh.seek(self.header0.seek_offset(guess-1))
                self.header0.fromfile(fh).time

        return guess


class GSBStreamReaderInfo(StreamReaderInfo):
    attr_names = list(StreamReaderInfo.attr_names)
    attr_names.insert(attr_names.index('readable'), 'bandwidth')
    attr_names.insert(attr_names.index('readable'), 'n_raw')
    attr_names.insert(attr_names.index('readable'), 'payload_nbytes')
    attr_names = tuple(attr_names)

    payload_nbytes = info_item(needs='_parent', doc=(
        'Number of bytes per payload (in each raw file).'))

    @info_item
    def frame0(self):
        """First frame from the file."""
        return self._parent._read_frame(0)

    # Bit of a hack, but the base reader one suffices here with
    # the frame0 override above.
    decodable = FileReaderInfo.decodable

    @info_item
    def file_info(self):
        """Information from timestamp file."""
        fh_ts_info = self._parent.fh_ts.info
        fh_ts_info.missing.pop('raw', None)
        return fh_ts_info

    @info_item(needs='shape')
    def bandwidth(self):
        """Bandwidth covered by the stream."""
        return (self.sample_rate * self.shape[-1]
                / (1 if self.complex_data else 2)).to(u.MHz)

    @info_item
    def n_raw(self):
        """Number of raw streams (per polarization)."""
        fh_raw = self._parent.fh_raw
        return len(fh_raw[0]) if isinstance(fh_raw, (list, tuple)) else 1

    @info_item(needs=('file_info', 'payload_nbytes', 'n_raw'), default=False)
    def consistent(self):
        """Whether timestamp and raw files are consistent in length."""
        pl_nbytes = self.payload_nbytes
        nchan = self._parent._unsliced_shape[-1]
        expected_size = int(((self.stop_time-self.start_time)
                             * self.sample_rate * nchan
                             * self.bps * (2 if self.complex_data else 1)
                             // (8 * self.n_raw)).to(u.one).round())
        fh_raw = self._parent.fh_raw
        if self.file_info.mode == 'rawdump':
            fh_raw = [[fh_raw]]

        msg = ''
        try:
            for pair in fh_raw:
                for fh in pair:
                    fs = file_size(fh)
                    if fs % pl_nbytes != 0 and 'non-integer' not in msg:
                        msg += ('raw file contains non-integer number ({}) '
                                'of payloads.'.format(fs / pl_nbytes))

                    consistent = fs >= expected_size
                    if not consistent:
                        emsg = 'raw file size smaller than expected.'
                        ratio = fs / expected_size
                        if len(pair) == 1 and 0.5 <= ratio < 0.6:
                            emsg = (emsg[:-1] + ' by {} factor of two. '
                                    'Are you missing the second raw file?'
                                    .format('a' if ratio == 0.5
                                            else 'about a'))
                        raise EOFError(emsg)

                    if fs > expected_size and 'more bytes' not in msg:
                        msg += 'raw file contains more bytes than expected.'
        finally:
            if msg:
                self.warnings['consistent'] = msg

        # As a final sanity check, try reading the final sample of the file.
        old_offset = self._parent.tell()
        try:
            self._parent.seek(-1, 2)
            self._parent.read(1)
        finally:
            self._parent.seek(old_offset)

        return True

    @info_item(needs='frame0', default=False)
    def readable(self):
        """Whether the file is readable and decodable."""
        self.checks['decodable'] = self.decodable
        self.checks['consistent'] = self.consistent
        return all(bool(v) for v in self.checks.values())



# Licensed under the GPLv3 - see LICENSE
import io
import pathlib

import pytest
import numpy as np
from astropy import units as u

from ...helpers import sequentialfile as sf
from ..header import HeaderParser, VLBIHeaderBase, four_word_struct
from ..base import (FileOpener,
                    FileBase, VLBIFileReaderBase,
                    StreamReaderBase, StreamWriterBase)


class BareHeader(VLBIHeaderBase):
    _struct = four_word_struct
    _header_parser = HeaderParser(
        (('x0_16_4', (0, 16, 4)),
         ('x0_31_1', (0, 31, 1, False)),
         ('x1_0_32', (1, 0, 32)),
         ('x2_0_64', (2, 0, 64, 1))))

    @property
    def payload_nbytes(self):
        return 10

    samples_per_frame = 10
    sample_shape = (2,)
    complex_data = False
    dtype = np.dtype('f4')


class BareFileReader(VLBIFileReaderBase):
    pass


class BareFileWriter(FileBase):
    pass


class BareStreamReader(StreamReaderBase):
    pass


class BareStreamWriter(StreamWriterBase):
    def __init__(self, fh_raw, header0, *,
                 squeeze=True, parrot='alife', **kwargs):
        if parrot == 'dead':
            raise ValueError('parrot is dead')
        super().__init__(fh_raw, header0, squeeze=squeeze, **kwargs)


class TestFileOpener:
    def setup_class(cls):
        cls.classes = {'rb': BareFileReader,
                       'wb': BareFileWriter,
                       'rs': BareStreamReader,
                       'ws': BareStreamWriter}
        cls.header_class = BareHeader
        cls.file_opener = FileOpener('Bare', classes=cls.classes,
                                     header_class=cls.header_class)
        cls.open = staticmethod(FileOpener.create(globals(), doc='extra'))

    def test_create_opener(self):
        assert self.open.__wrapped__.__func__ is FileOpener.__call__
        assert 'Open Bare file(s) for reading or writing.' in self.open.__doc__
        assert self.open.__doc__.endswith('extra')

    def test_create_opener_wrong_ns(self):
        with pytest.raises(ValueError, match='does not contain'):
            FileOpener.create(locals(), doc='extra')

    def test_methods(self):
        assert self.file_opener.is_name('abc')
        assert not self.file_opener.is_name(['a', 'b'])
        assert self.file_opener.is_template('{abc}')
        assert self.file_opener.is_sequence('{abc}')
        assert not self.file_opener.is_template('abc')

    def test_get_type_custom_indexer(self):
        class MySeq:
            def __getitem__(self, index):
                return f'{index:07d}.bare'

        my_seq = MySeq()
        assert my_seq[10] == '0000010.bare'
        assert self.file_opener.get_type(my_seq) == 'sequence'
        assert not self.file_opener.is_name(my_seq)
        assert self.file_opener.is_sequence(my_seq)
        assert not self.file_opener.is_template(my_seq)

    @pytest.mark.parametrize('kwargs,expected,left', [
        (dict(), BareHeader.fromvalues(), dict()),
        (dict(x0_16_4=0), BareHeader.fromvalues(x0_16_4=0), dict()),
        (dict(squeeze=True), BareHeader.fromvalues(), dict(squeeze=True)),
        (dict(parrot='dead'), BareHeader.fromvalues(), dict(parrot='dead'))])
    def test_get_header0(self, kwargs, expected, left):
        header0 = self.file_opener.get_header0(kwargs)
        assert header0 == expected
        assert kwargs == left

    def test_get_header0_failure(self):
        with pytest.raises(AttributeError):
            self.file_opener.get_header0(dict(payload_nbytes=2))

    def test_get_fns_with_header_failure(self):
        # But still should be able to get a fns.
        fns = self.file_opener.get_fns('{file_nr}_{bla}.bare', 'rs',
                                       dict(payload_nbytes=2, bla='aha'))
        assert fns[1] == '1_aha.bare'

    @pytest.mark.parametrize('path_type', [str, None, pathlib.Path])
    def test_binary_name(self, path_type, tmpdir):
        path = tmpdir.join('test.bare')
        if path_type:
            path = path_type(path)
        name = str(path)

        with self.file_opener(path, 'wb') as fw:
            assert isinstance(fw, BareFileWriter)
            assert fw.fh_raw.name == name
            fw.write(b'abcde')

        with self.file_opener(path, 'rb') as fr:
            assert isinstance(fr, BareFileReader)
            assert fr.fh_raw.name == name
            assert fr.read() == b'abcde'

    @pytest.mark.parametrize('mode', ['rb', 'rs'])
    def test_file_not_found(self, mode):
        with pytest.raises(FileNotFoundError):
            self.file_opener('does_not_exist', mode)

    def test_binary_fh(self, tmpdir):
        # Also flip mode, and use constructed open
        name = str(tmpdir.join('test.bare'))
        with io.open(name, 'wb') as fh:
            with self.open(fh, 'bw') as fw:
                assert isinstance(fw, BareFileWriter)
                assert fw.fh_raw is fh
                fw.write(b'abcde')

        with io.open(name, 'rb') as fh:
            with self.open(fh, 'br') as fr:
                assert isinstance(fr, BareFileReader)
                assert fr.fh_raw is fh
                assert fr.read() == b'abcde'

    def test_stream_template(self, tmpdir):
        template = str(tmpdir.join('{x2_0_64:03d}_{file_nr:02d}.bare'))
        header0 = BareHeader.fromvalues()
        with self.open(template, 'ws', header0=header0,
                       sample_rate=10*u.Hz) as fw:
            assert isinstance(fw, BareStreamWriter)
            assert isinstance(fw.fh_raw, sf.SequentialFileWriter)
            assert isinstance(fw.fh_raw.files, sf.FileNameSequencer)
            assert fw.fh_raw.files[0].endswith('001_00.bare')

        with self.open(template, 'w', x2_0_64=4,
                       sample_rate=10*u.Hz, parrot='life') as fw:
            assert fw.fh_raw.files[0].endswith('004_00.bare')

        with pytest.raises(TypeError):
            self.open(template, 'ws', coocoo=10,
                      sample_rate=10*u.Hz)

    @pytest.mark.parametrize('module,doc', [
        (None, None),
        (None, 'new_doc'),
        (__name__, 'new_doc')])
    def test_wrapped(self, module, doc):
        open = self.file_opener.wrapped(module, doc)
        if doc is None:
            assert open.__doc__ is self.file_opener.__call__.__doc__
        else:
            assert open.__doc__ is doc

        if module is None:
            assert open.__module__ is self.file_opener.__call__.__module__
        else:
            assert open.__module__ is module

    def test_invalid_name(self):
        with pytest.raises(ValueError, match='not understood'):
            self.file_opener(0, 'wb')

    def test_invalid_mode(self):
        name = 'not_needed'
        with pytest.raises(ValueError, match='invalid mode'):
            self.file_opener(name, 'xb')

        with pytest.raises(ValueError, match='invalid mode'):
            self.file_opener(name, 's')

    def test_binary_no_template(self):
        with pytest.raises(ValueError, match='does not support'):
            self.file_opener('{file_nr}.bare', 'wb')



# Licensed under the GPLv3 - see LICENSE
import warnings
from collections import namedtuple

import numpy as np
import astropy.units as u
from astropy.utils import lazyproperty

from ..base.base import (
    FileBase, VLBIFileReaderBase,
    VLBIStreamReaderBase, StreamWriterBase,
    FileOpener, FileInfo, HeaderNotFoundError)
from .header import VDIFHeader
from .payload import VDIFPayload
from .frame import VDIFFrame, VDIFFrameSet
from .file_info import VDIFFileReaderInfo


__all__ = ['VDIFFileReader', 'VDIFFileWriter',
           'VDIFStreamBase', 'VDIFStreamReader', 'VDIFStreamWriter',
           'open', 'info']

# Check code on 2015-MAY-30
# 00000000  77 2c db 00 00 00 00 1c  75 02 00 20 fc ff 01 04  # header 0 - 3
# 00000010  10 00 80 03 ed fe ab ac  00 00 40 33 83 15 03 f2  # header 4 - 7
# 00000020  2a 0a 7c 43 8b 69 9d 59  cb 99 6d 9a 99 96 5d 67  # data 0 - 3
# NOTE: thread_id = 1
# 2a = 00 10 10 10 = (lsb first) 1,  1,  1, -3
# 0a = 00 00 10 10 =             1,  1, -3, -3
# 7c = 01 11 11 00 =            -3,  3,  3, -1
# m5d evn/Fd/GP052D_FD_No0006.m5a VDIF_5000-512-1-2 100
# Mark5 stream: 0x16cd140
#   stream = File-1/1=evn/Fd/GP052D_FD_No0006.m5a
#   format = VDIF_5000-512-1-2 = 3
#   start mjd/sec = 56824 21367.000000000
#   frame duration = 78125.00 ns
#   framenum = 0
#   sample rate = 256000000 Hz
#   offset = 0
#   framebytes = 5032 bytes
#   datasize = 5000 bytes
#   sample granularity = 4
#   frame granularity = 1
#   gframens = 78125
#   payload offset = 32
#   read position = 0
#   data window size = 1048576 bytes
#  1  1  1 -3  1  1 -3 -3 -3  3  3 -1  -> OK
# fh = vdif.open('evn/Fd/GP052D_FD_No0006.m5a', 'rb')
# fs = fh.read_frameset()
# fs.data.astype(int)[1, :12, 0]  # thread id = 1!!
# -> array([ 1,  1,  1, -3,  1,  1, -3, -3, -3,  3,  3, -1])  -> OK
# Also, next frame (thread #3)
# m5d evn/Fd/GP052D_FD_No0006.m5a VDIF_5000-512-1-2 12 5032
# -1  1 -1  1 -3 -1  3 -1  3 -3  1  3
# fs.data.astype(int)[3, :12, 0]
# -> array([-1,  1, -1,  1, -3, -1,  3, -1,  3, -3,  1,  3])
# And first thread #0
# m5d evn/Fd/GP052D_FD_No0006.m5a VDIF_5000-512-1-2 12 20128
# -1 -1  3 -1  1 -1  3 -1  1  3 -1  1
# fs.data.astype(int)[0, :12, 0]
# -> array([-1, -1,  3, -1,  1, -1,  3, -1,  1,  3, -1,  1])
# sanity check that we can read 12 samples with stream reader
# fh.close()
# fh = vdif.open('evn/Fd/GP052D_FD_No0006.m5a', 'rs')
# fh.read(12).astype(int)[:, 0]
# -> array([-1, -1,  3, -1,  1, -1,  3, -1,  1,  3, -1,  1])


class VDIFFileReader(VLBIFileReaderBase):
    """Simple reader for VDIF files.

    Wraps a binary filehandle, providing methods to help interpret the data,
    such as `read_frame`, `read_frameset` and `get_frame_rate`.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw binary data file.

    """
    info = VDIFFileReaderInfo()

    def read_header(self, edv=None, verify=True):
        """Read a single header from the file.

        Parameters
        ----------
        edv : int, False, or None, optional
            Extended data version.  If `False`, a legacy header is used.
            If `None` (default), it is determined from the header.  (Given it
            explicitly is mostly useful for a slight speed-up.)
        verify : bool, optional
            Whether to do basic verification of integrity.  Default: `True`.

        Returns
        -------
        header : `~baseband.vdif.VDIFHeader`
        """
        return VDIFHeader.fromfile(self.fh_raw, edv=edv, verify=verify)

    def read_frame(self, edv=None, verify=True):
        """Read a single frame (header plus payload).

        Parameters
        ----------
        edv : int, optional
            The expected extended data version for the VDIF Header.  If `None`,
            use that of the first frame.  (Passing it in slightly improves file
            integrity checking.)
        verify : bool, optional
            Whether to do basic checks of frame integrity.  Default: `True`.

        Returns
        -------
        frame : `~baseband.vdif.VDIFFrame`
            With ``.header`` and ``.data`` properties that return the
            :class:`~baseband.vdif.VDIFHeader` and data encoded in the frame,
            respectively.
        """
        return VDIFFrame.fromfile(self.fh_raw, edv=edv, verify=verify)

    def read_frameset(self, thread_ids=None, edv=None, verify=True):
        """Read a single frame (header plus payload).

        Parameters
        ----------
        thread_ids : list, optional
            The thread ids that should be read.  If `None` (default), read all
            threads.
        edv : int, optional
            The expected extended data version for the VDIF Header.  If `None`,
            use that of the first frame.  (Passing it in slightly improves file
            integrity checking.)
        verify : bool, optional
            Whether to do basic checks of frame integrity.  Default: `True`.

        Returns
        -------
        frameset : :class:`~baseband.vdif.VDIFFrameSet`
            With ``.headers`` and ``.data`` properties that return a list of
            :class:`~baseband.vdif.VDIFHeader` and the data encoded in the
            frame set, respectively.
        """
        return VDIFFrameSet.fromfile(self.fh_raw, thread_ids, edv=edv,
                                     verify=verify)

    def get_frame_rate(self):
        """Determine the number of frames per second.

        This method first tries to determine the frame rate by looking for
        the highest frame number in the first second of data.  If that fails,
        it attempts to extract the sample rate from the header.

        Returns
        -------
        frame_rate : `~astropy.units.Quantity`
            Frames per second.
        """
        try:
            return super().get_frame_rate()
        except Exception as exc:
            with self.temporary_offset(0):
                try:
                    header = self.read_header()
                    return (header.sample_rate
                            / header.samples_per_frame).to(u.Hz).round()
                except Exception:
                    pass
            raise exc

    def get_thread_ids(self, check=2):
        """Determine the number of threads in the VDIF file.

        The file is presumed to be positioned at the start of a header.
        Usually, it suffices to just seek to the start of the file, but
        if not, use `~baseband.vdif.base.VDIFFileReader.find_header`.

        Parameters
        ----------
        check : int, optional
            Number of extra frames to check.  Frame sets are scanned until
            the number of thread IDs found no longer increases for ``check``
            frames.

        Returns
        -------
        thread_ids : list
            Sorted list of all thread ids encountered in the frames scanned.
        """
        with self.temporary_offset():
            header = header0 = self.read_header()
            try:
                thread_ids = set()
                n_check = 1
                while n_check > 0:
                    frame_nr = header['frame_nr']
                    n_thread = len(thread_ids)
                    while header['frame_nr'] == frame_nr:
                        thread_ids.add(header['thread_id'])
                        self.seek(header.payload_nbytes, 1)
                        header = self.read_header(edv=header0.edv)
                        assert header0.same_stream(header)

                    if len(thread_ids) > n_thread:
                        n_check = check
                    else:
                        n_check -= 1
            except EOFError:
                # Hack: let through very short files (like our samples).
                if self.seek(0, 2) > (check * len(thread_ids)
                                      * header0.frame_nbytes):
                    raise

        return sorted(thread_ids)

    def find_header(self, pattern=None, *, edv=None, mask=None,
                    frame_nbytes=None, offset=0,
                    forward=True, maximum=None, check=1):
        """Find the nearest header from the current position.

        Search for a valid header at a given position which is consistent with
        ``pattern`` and/or with a header a frame size ahead.  Note
        that the search is much slower if no pattern is given, as at every
        position it is tried to read a header, and then check for another one
        one frame ahead.  It helps to pass in ``edv`` and ``frame_nbytes``
        (if known).

        If successful, the file pointer is left at the start of the header.

        Parameters
        ----------
        pattern : `~baseband.vdif.VDIFHeader`, array of byte, or compatible
            If given, used for a direct search.
        edv : int
            EDV of the header, used if ``pattern`` is not given.
        mask : array of byte, bytes, iterable of int, string or int
            Bit mask for the pattern, with 1 indicating a given bit will
            be used the comparison.  Only used with ``pattern`` and not
            needed if ``pattern`` is a header.
        frame_nbytes : int, optional
            Frame size in bytes.  Defaults to the frame size in any header
            passed in.
        offset : int, optional
            Offset from the frame start that the pattern occurs.  Any
            offsets inferred from masked entries are added to this (hence,
            no offset needed when a header is passed in as ``pattern``,
            nor is an offset needed for a full search).
        forward : bool, optional
            Seek forward if `True` (default), backward if `False`.
        maximum : int, optional
            Maximum number of bytes to search away from the present location.
            Default: search twice the frame size if given, otherwise 10000
            (extra bytes to avoid partial patterns will be added).
            Use 0 to check only at the current position.
        check : int or tuple of int, optional
            Frame offsets where another header should be present.
            Default: 1, i.e., a sync pattern should be present one
            frame after the one found (independent of ``forward``),
            thus helping to guarantee the frame is not corrupted.

        Returns
        -------
        header : :class:`~baseband.vdif.VDIFHeader`
            Retrieved VDIF header.

        Raises
        ------
        ~baseband.base.base.HeaderNotFoundError
            If no header could be located.
        AssertionError
            If the header did not pass verification.
        """
        if pattern is not None:
            locations = self.locate_frames(
                pattern, mask=mask, frame_nbytes=frame_nbytes, offset=offset,
                forward=forward, maximum=maximum, check=check)
            if not locations:
                raise HeaderNotFoundError('could not locate a a nearby frame.')
            self.seek(locations[0])
            with self.temporary_offset():
                return self.read_header(edv=getattr(pattern, 'edv', None))

        # Try reading headers at a set of locations.
        if maximum is None:
            maximum = 10000 if frame_nbytes is None else 2 * frame_nbytes

        file_pos = self.tell()
        # Generate file pointer positions to test.
        if forward:
            iterate = range(file_pos, file_pos+maximum+1)
        else:
            iterate = range(file_pos, max(file_pos-maximum-1, -1), -1)
        # Loop over all of them to try to find the frame marker.
        for frame in iterate:
            self.seek(frame)
            try:
                header = self.read_header(edv=edv)
            except Exception:
                continue

            if (frame_nbytes is not None
                    and frame_nbytes != header.frame_nbytes):
                continue

            # Possible hit!  Try if there are other headers right around.
            self.seek(frame)
            try:
                return self.find_header(header, maximum=0, check=check)
            except Exception:
                continue

        self.seek(file_pos)
        raise HeaderNotFoundError("could not locate a nearby header.")


class VDIFFileWriter(FileBase):
    """Simple writer for VDIF files.

    Adds `write_frame` and `write_frameset` methods to the basic VLBI
    binary file wrapper.
    """

    def write_frame(self, data, header=None, **kwargs):
        """Write a single frame (header plus payload).

        Parameters
        ----------
        data : `~numpy.ndarray` or `~baseband.vdif.VDIFFrame`
            If an array, a ``header`` should be given, which will be used to
            get the information needed to encode the array, and to construct
            the VDIF frame.
        header : `~baseband.vdif.VDIFHeader`
            Can instead give keyword arguments to construct a header.  Ignored
            if ``data`` is a `~baseband.vdif.VDIFFrame` instance.
        **kwargs
            If ``header`` is not given, these are used to initialize one.
        """
        if not isinstance(data, VDIFFrame):
            data = VDIFFrame.fromdata(data, header, **kwargs)
        return data.tofile(self.fh_raw)

    def write_frameset(self, data, header=None, **kwargs):
        """Write a single frame set (headers plus payloads).

        Parameters
        ----------
        data : `~numpy.ndarray` or :class:`~baseband.vdif.VDIFFrameSet`
            If an array, a header should be given, which will be used to
            get the information needed to encode the array, and to construct
            the VDIF frame set.
        header : :class:`~baseband.vdif.VDIFHeader`, list of same
            Can instead give keyword arguments to construct a header.  Ignored
            if ``data`` is a :class:`~baseband.vdif.VDIFFrameSet` instance.
            If a list, should have a length matching the number of threads in
            ``data``; if a single header, ``thread_ids`` corresponding
            to the number of threads are generated automatically.
        **kwargs
            If ``header`` is not given, these are used to initialize one.
        """
        if not isinstance(data, VDIFFrameSet):
            data = VDIFFrameSet.fromdata(data, header, **kwargs)
        return data.tofile(self.fh_raw)


class VDIFStreamBase:
    """Provides sample shape maker and fast time and index getting/setting."""

    _sample_shape_maker = namedtuple('SampleShape', 'nthread, nchan')

    def _get_time(self, header):
        """Get time from a header.

        This passes on sample rate, since not all VDIF headers can calculate
        it.
        """
        return header.get_time(frame_rate=self._frame_rate)

    def _set_time(self, header, time):
        """Set time in a header.

        This passes on sample rate, which not all VDIF headers can calculate.
        """
        header.update(time=time, frame_rate=self._frame_rate)

    def _get_index(self, header):
        # Override to avoid explicit time calculations.
        return int(round((header['seconds'] - self.header0['seconds'])
                         * self._frame_rate.to_value(u.Hz)
                         + header['frame_nr'] - self.header0['frame_nr']))

    def _set_index(self, header, index):
        # Override to avoid explicit time calculations.
        dt, frame_nr = divmod(index + self.header0['frame_nr'],
                              int(round(self._frame_rate.to_value(u.Hz))))
        seconds = self.header0['seconds'] + dt
        header['seconds'] = seconds
        header['frame_nr'] = frame_nr

    def __repr__(self):
        return ("<{s.__class__.__name__} name={s.name} offset={s.offset}\n"
                "    sample_rate={s.sample_rate},"
                " samples_per_frame={s.samples_per_frame},\n"
                "    sample_shape={s.sample_shape},\n"
                "    bps={s.bps}, complex_data={s.complex_data},"
                " edv={h.edv}, station={h.station},\n"
                "    {sub}start_time={s.start_time}>"
                .format(s=self, h=self.header0,
                        sub=('subset={0}, '.format(self.subset)
                             if self.subset else '')))


class VDIFStreamReader(VDIFStreamBase, VLBIStreamReaderBase):
    """VLBI VDIF format reader.

    Allows access to a VDIF file as a continuous series of samples.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw VDIF stream.
    sample_rate : `~astropy.units.Quantity`, optional
        Number of complete samples per second, i.e. the rate at which each
        channel in each thread is sampled.  If `None` (default), will be
        inferred from the header or by scanning one second of the file.
    squeeze : bool, optional
        If `True` (default), remove any dimensions of length unity from
        decoded data.
    subset : indexing object or tuple of objects, optional
        Specific components of the complete sample to decode (after possible
        squeezing).  If a single indexing object is passed, it selects threads.
        If a tuple is passed, the first selects threads and the second selects
        channels.  If the tuple is empty (default), all components are read.
    fill_value : float or complex, optional
        Value to use for invalid or missing data. Default: 0.
    verify : bool, optional
        Whether to do basic checks of frame integrity when reading.  The first
        frameset of the stream is always checked.  Default: `True`.
    """

    def __init__(self, fh_raw, sample_rate=None, squeeze=True, subset=(),
                 fill_value=0., verify='fix'):
        fh_raw = VDIFFileReader(fh_raw)
        # We read the very first header, hoping this is the right one
        # (in some VLBA files not all the headers have the right time).
        header0 = fh_raw.read_header()
        # Next, we determine how many threads there are, and use those
        # to calculate the frameset size.  We on purpose do *not* just read
        # a frame set, since sometimes the first one is short (see gh-359).
        fh_raw.seek(0)
        thread_ids = fh_raw.get_thread_ids()
        nthread = len(thread_ids)
        super().__init__(
            fh_raw, header0, sample_rate=sample_rate,
            sample_shape=(nthread, header0.nchan), squeeze=squeeze,
            subset=subset, fill_value=fill_value, verify=verify)
        self._raw_offsets.frame_nbytes *= nthread

        # Check whether we are reading only some threads.  This is somewhat
        # messy since normally we apply the whole subset to the whole data,
        # but here we need to split it up in the part that selects specific
        # threads, which we use to selectively read, and the rest, which we
        # do post-decoding.
        if self.subset and (nthread > 1 or not self.squeeze):
            # Select the thread ids we want using first part of subset.
            thread_ids = np.array(thread_ids)[self.subset[0]]
            # Use squeese in case subset[0] uses broadcasting, and
            # atleast_1d to ensure single threads get upgraded to a list,
            # which is needed by the VDIFFrameSet reader.
            self._thread_ids = np.atleast_1d(thread_ids.squeeze()).tolist()
            # Since we have subset the threads already, we now need to
            # determine a new subset that takes this into account.
            if thread_ids.shape == ():
                # If we indexed with a scalar, we're meant to remove that
                # dimension. If we squeeze, this happens automatically, but if
                # not, we need to do it explicitly (FrameSet does not do it).
                new_subset0 = () if self.squeeze else (0,)
            elif len(self._thread_ids) == 1 and self.squeeze:
                # If we want a single remaining thread, undo the squeeze.
                new_subset0 = (np.newaxis,)
            else:
                # Just pass on multiple threads or unsqueezed single ones.
                new_subset0 = (slice(None),)

            self._frameset_subset = new_subset0 + self.subset[1:]
        else:
            # We either have no subset or we have a single thread that
            # will be squeezed away, so the subset is fine as is.
            self._frameset_subset = self.subset
            self._thread_ids = thread_ids

    @lazyproperty
    def _last_header(self):
        """Last header of the file."""
        # Go to end of file.
        maximum = 2 * self._raw_offsets.frame_nbytes
        with self.fh_raw.temporary_offset(
                -self.header0.frame_nbytes, 2) as fh_raw:
            # Find first header with same thread_id going backward.
            locations = fh_raw.locate_frames(self.header0, forward=False,
                                             maximum=maximum, check=(-1, 1))
            for location in locations:
                fh_raw.seek(location)
                try:
                    header = fh_raw.read_header(edv=self.header0.edv)
                except Exception:  # pragma: no cover
                    # If reading fails, just try the next one -- we would
                    # use loose a bit of the end of the file.  Would require
                    # an EDV verify stricter than what locate_frames uses.
                    continue

                if header['thread_id'] == self.header0['thread_id']:
                    return header

            raise HeaderNotFoundError(
                "corrupt VDIF? No thread_id={0} frame in last {1} bytes."
                .format(self.header0['thread_id'], maximum))

    def _squeeze_and_subset(self, data):
        # Overwrite VLBIStreamReaderBase version, since the threads part of
        # subset has already been used.
        if self.squeeze:
            data = data.reshape(data.shape[:1]
                                + tuple(sh for sh in data.shape[1:] if sh > 1))
        if self._frameset_subset:
            data = data[(slice(None),) + self._frameset_subset]

        return data

    # Overrides to deal with framesets instead of frames.
    def _fh_raw_read_frame(self):
        return self.fh_raw.read_frameset(self._thread_ids,
                                         edv=self.header0.edv,
                                         verify=self.verify)

    def _bad_frame(self, index, frameset, exc):
        # Duplication of base class, but able to deal with missing
        # frames inside a frame set.
        if (frameset is not None and self._get_index(frameset) == index
                and index == self._get_index(self._last_header)):
            # If we got an exception because we're trying to read beyond the
            # last frame, the frame is almost certainly OK, so keep it.
            return frameset

        if self.verify != 'fix':
            raise exc

        # If the frameset does contain the right number of frames, but all
        # are invalid, assume not just the data but also the frame number
        # and seconds might be wrong, i.e., just proceed with it.
        # TODO: make this an option for a specific type of fixing!
        if (frameset is not None
                and len(frameset.frames) == len(self._thread_ids)
                and not any(frame.valid for frame in frameset.frames)):
            return frameset

        msg = 'problem loading frame set {}.'.format(index)

        # Where should we be?
        raw_offset = self._seek_frame(index)

        # See if we're in the right place.  First ensure we have a header.
        # Here, it is more important that it is a good one than that we go
        # too far, so we insist on two consistent frames after it, as well
        # as a good one before to guard against corruption of the start of
        # the VDIF header.
        self.fh_raw.seek(raw_offset)
        try:
            header = self.fh_raw.find_header(
                self.header0, forward=True, check=(-1, 1, 2),
                maximum=3*self.header0.frame_nbytes)
        except HeaderNotFoundError:
            exc.args += (msg + ' Cannot find header nearby.',)
            raise exc

        # Don't yet know how to deal with excess data.
        header_index = self._get_index(header)
        if header_index < index:
            exc.args += (msg + ' There appears to be excess data.',)
            raise exc

        # Go backward until we find previous frame, storing offsets
        # as we go.  We again increase the maximum since we may need
        # to jump over a bad bit.  We slightly relax our search pattern.
        while header_index >= index:
            raw_pos = self.fh_raw.tell()
            header1 = header
            header1_index = header_index
            if raw_pos <= 0:
                break

            self.fh_raw.seek(-1, 1)
            try:
                header = self.fh_raw.find_header(
                    self.header0, forward=False,
                    maximum=4*self.header0.frame_nbytes,
                    check=(-1, 1))
            except HeaderNotFoundError:
                exc.args += (msg + ' Could not find previous index.',)
                raise exc

            header_index = self._get_index(header)
            if header_index < header1_index:
                # While we are at it: if we pass an index boundary,
                # update the list of known indices.
                self._raw_offsets[header1_index] = raw_pos

        # Move back to position of last good header (header1).
        self.fh_raw.seek(raw_pos)

        # Create the header we will use below for constructing the
        # frameset.  Usually, this is guaranteed to be from this set,
        # but we also use it to create a new header for a completely
        # messed up frameset below.  It is copied to make it mutable
        # without any risk of messing up possibly memory mapped data.
        header = header1.copy()

        if header1_index > index:
            # Ouch, whole frame set missing!
            msg += ' The frame set seems to be missing altogether.'
            # Set up to construct a complete missing frame
            # (after the very long else clause).
            # TODO: just use the writer's _make_frame??
            frames = {}
            self._set_index(header, index)
        else:
            assert header1_index == index, \
                'at this point, we should have a good header.'
            # This header is the first one of its set.
            if raw_pos != raw_offset:
                msg += ' Stream off by {0} bytes.'.format(raw_offset
                                                          - raw_pos)
                # Above, we should have added information about
                # this index in our offset table.
                assert raw_pos == self._raw_offsets[index]

            # Try again to read it, however many threads there are.
            # TODO: this somewhat duplicates FrameSet.fromfile; possibly
            #       move code there.
            # TODO: Or keep track of header locations above.
            # TODO: remove limitation that threads need to be together.
            frames = {}
            previous = False
            frame_nr = header1['frame_nr']
            while True:
                raw_pos = self.fh_raw.tell()
                try:
                    frame = self.fh_raw.read_frame(edv=self.header0.edv)
                    assert header.same_stream(frame.header)
                    # Check seconds as well, as a sanity check this is
                    # a real header. (We do allow at this point it to be
                    # the next frame, hence the second can increase by 1.
                    assert 0 <= (frame['seconds'] - header['seconds']) <= 1
                except EOFError:
                    # End of file while reading a frame; we're done here.
                    next_header = None
                    break
                except AssertionError:
                    # Frame is not OK.
                    assert previous is not False, \
                        ('first frame should be readable if fully on disk,'
                         ' since we found one correct header.')

                    # Go back to after previous payload and try finding
                    # next header.  It can be before where we tried above,
                    # if some bytes in the previous payload were missing.
                    self.fh_raw.seek(raw_pos - header.payload_nbytes)
                    try:
                        next_header = self.fh_raw.find_header(self.header0)
                        # But sometimes a header is re-found even when
                        # there isn't really one (e.g., because one of the
                        # first bytes, defining seconds, is missing).
                        # Don't ever retry the same one!
                        if self.fh_raw.tell() == raw_pos:
                            self.fh_raw.seek(1, 1)
                            next_header = self.fh_raw.find_header(self.header0)
                    except HeaderNotFoundError:
                        # If no header was found, give up.  The previous frame
                        # was likely bad too, so delete it.
                        if previous is not None:
                            del frames[previous]
                        next_header = None
                        break

                    # If the next header is not exactly a frame away from
                    # where we were trying to read, the previous frame was
                    # likely bad, so discard it.
                    if self.fh_raw.tell() != raw_pos + header.frame_nbytes:
                        if previous is not None:
                            del frames[previous]
                        previous = None

                    # Stop if the next header is from a different frame.
                    if next_header['frame_nr'] != frame_nr:
                        break

                else:
                    # Successfully read frame.  If not of the requested
                    # set, rewind and break out of the loop.
                    if frame['frame_nr'] != frame_nr:
                        next_header = frame.header
                        self.fh_raw.seek(raw_pos)
                        break

                    # Do we have a good frame, giving a new thread?
                    previous = frame['thread_id']
                    if previous in frames:
                        msg += (' Duplicate thread {0} found; discarding.'
                                .format(previous))
                        del frames[previous]
                    else:
                        # Looks like it, though may still be discarded
                        # if the next frame is not readable.
                        frames[previous] = frame

            # If the next header is of the next frame, set up the raw
            # offset (which likely will be needed, saving some time).
            if (next_header is not None
                    and self._get_index(next_header) == index + 1):
                self._raw_offsets[index+1] = self.fh_raw.tell()

        # Create invalid frame template,
        invalid_payload = VDIFPayload(
            np.zeros(header.payload_nbytes // 4, '<u4'), header)
        invalid_frame = VDIFFrame(header, invalid_payload, valid=False)

        frame_list = []
        missing = []
        for thread in self._thread_ids:
            if thread in frames:
                frame_list.append(frames[thread])
            else:
                missing.append(thread)
                invalid_frame.header['thread_id'] = thread
                frame_list.append(invalid_frame)

        if missing:
            if frames == {}:
                msg += ' All threads set to invalid.'
            else:
                msg += (' Thread(s) {0} missing; set to invalid.'
                        .format(missing))

        warnings.warn(msg)
        frameset = VDIFFrameSet(frame_list)
        return frameset


class VDIFStreamWriter(VDIFStreamBase, StreamWriterBase):
    """VLBI VDIF format writer.

    Encodes and writes sequences of samples to file.

    Parameters
    ----------
    fh_raw : filehandle
        Which will write filled sets of frames to storage.
    header0 : :class:`~baseband.vdif.VDIFHeader`
        Header for the first frame, holding time information, etc.
    sample_rate : `~astropy.units.Quantity`
        Number of complete samples per second, i.e. the rate at which each
        channel in each thread is sampled.  For EDV 1 and 3, can
        alternatively set ``sample_rate`` within the header.
    nthread : int, optional
        Number of threads (e.g., 2 for 2 polarisations).  Default: 1.
    squeeze : bool, optional
        If `True` (default), `write` accepts squeezed arrays as input, and
        adds any dimensions of length unity.
    """

    def __init__(self, fh_raw, header0=None, sample_rate=None, nthread=1,
                 squeeze=True):
        fh_raw = VDIFFileWriter(fh_raw)
        # Get header sample rate
        try:
            header_sample_rate = header0.sample_rate
        except AttributeError:
            header_sample_rate = None

        if sample_rate is None:
            if header_sample_rate is None:
                raise ValueError("the sample rate must be passed either "
                                 "explicitly, or through the header if it "
                                 "can be stored there.")

            sample_rate = header_sample_rate
        elif header_sample_rate is not None:
            assert sample_rate == header_sample_rate, (
                'sample_rate on header inconsistent with that passed in.')

        super().__init__(fh_raw, header0, sample_rate=sample_rate,
                         sample_shape=(nthread, header0.nchan),
                         squeeze=squeeze)

        self._frame = VDIFFrameSet.fromdata(
            np.zeros((self.samples_per_frame, nthread, header0.nchan),
                     dtype=np.complex64 if self.complex_data else np.float32),
            self.header0)


open = FileOpener.create(globals(), doc="""
--- For reading a stream : (see :class:`~baseband.vdif.base.VDIFStreamReader`)

sample_rate : `~astropy.units.Quantity`, optional
    Number of complete samples per second, i.e. the rate at which each channel
    in each thread is sampled.  If `None` (default), will be inferred from the
    header or by scanning one second of the file.
squeeze : bool, optional
    If `True` (default), remove any dimensions of length unity from
    decoded data.
subset : indexing object or tuple of objects, optional
    Specific components of the complete sample to decode (after possible
    squeezing).  If a single indexing object is passed, it selects threads.
    If a tuple is passed, the first selects threads and the second selects
    channels.  If the tuple is empty (default), all components are read.
fill_value : float or complex, optional
    Value to use for invalid or missing data. Default: 0.
verify : bool, optional
    Whether to do basic checks of frame integrity when reading.  The first
    frameset of the stream is always checked.  Default: `True`.

--- For writing a stream : (see :class:`~baseband.vdif.base.VDIFStreamWriter`)

header0 : `~baseband.vdif.VDIFHeader`
    Header for the first frame, holding time information, etc.  Can instead
    give keyword arguments to construct a header (see ``**kwargs``).
sample_rate : `~astropy.units.Quantity`
    Number of complete samples per second, i.e. the rate at which each
    channel in each thread is sampled.  For EDV 1 and 3, can alternatively set
    ``sample_rate`` within the header.
nthread : int, optional
    Number of threads (e.g., 2 for 2 polarisations).  Default: 1.
squeeze : bool, optional
    If `True` (default), writer accepts squeezed arrays as input, and adds any
    dimensions of length unity.
file_size : int or None, optional
    When writing to a sequence of files, the maximum size of one file in bytes.
    If `None` (default), the file size is unlimited, and only the first
    file will be written to.
**kwargs
    If no header is given, an attempt is made to construct one from these.
    For a standard header, this would include the following.

--- Header keywords : (see :meth:`~baseband.vdif.VDIFHeader.fromvalues`)

time : `~astropy.time.Time`
    Start time of the file.  Can instead pass on ``ref_epoch`` and
    ``seconds``.
nchan : int, optional
    Number of channels (default: 1).  Note: different numbers of channels
    per thread is not supported.
complex_data : bool, optional
    Whether data are complex.  Default: `False`.
bps : int, optional
    Bits per elementary sample, i.e. per real or imaginary component for
    complex data.  Default: 1.
samples_per_frame : int
    Number of complete samples per frame.  Can alternatively use
    ``frame_length``, the number of 8-byte words for header plus payload.
    For some EDV, this number is fixed (e.g., ``frame_length=629`` for
    ``edv=3``, which corresponds to 20000 real 2-bit samples per frame).
station : 2 characters, optional
    Station ID.  Can also be an unsigned 2-byte integer.  Default: 0.
edv : {`False`, 0, 1, 2, 3, 4, 0xab}
    Extended Data Version.

Notes
-----
One can also pass to ``name`` a list, tuple, or subclass of
`~baseband.helpers.sequentialfile.FileNameSequencer`.  For writing to multiple
files, the ``file_size`` keyword must be passed or only the first file will be
written to.  One may also pass in a `~baseband.helpers.sequentialfile` object
(opened in 'rb' mode for reading or 'w+b' for writing), though for typical use
cases it is practically identical to passing in a list or template.
""")


info = FileInfo.create(globals())



# Licensed under the GPLv3 - see LICENSE
"""Payload for DADA format."""
from collections import namedtuple

import numpy as np

from ..base.payload import PayloadBase


__all__ = ['DADAPayload']


def decode_8bit(words):
    return words.view(np.int8, np.ndarray).astype(np.float32)


def encode_8bit(values):
    return np.clip(np.rint(values), -128, 127).astype(np.int8)


class DADAPayload(PayloadBase):
    """Container for decoding and encoding DADA payloads.

    Parameters
    ----------
    words : `~numpy.ndarray`
        Array containg LSB unsigned words (with the right size) that
        encode the payload.
    header : `~baseband.dada.DADAHeader`
        Header that provides information about how the payload is encoded.
        If not given, the following arguments have to be passed in.
    bps : int, optional
        Number of bits per sample part (i.e., per channel and per real or
        imaginary component).  Default: 8.
    sample_shape : tuple, optional
        Shape of the samples; e.g., (nchan,).  Default: ().
    complex_data : bool, optional
        Whether data are complex.  Default: `False`.
    """
    _decoders = {
        8: decode_8bit}
    _encoders = {
        8: encode_8bit}
    _memmap = True
    _sample_shape_maker = namedtuple('SampleShape', 'npol, nchan')



# Licensed under the GPLv3 - see LICENSE
import pytest

from ..offsets import RawOffsets


class TestRawOffsets:
    @pytest.mark.parametrize('frame_nbytes', [0, 10, 100])
    def test_plain(self, frame_nbytes):
        ro = RawOffsets(frame_nbytes=frame_nbytes)
        assert ro[0] == 0
        assert ro[1] == frame_nbytes
        assert ro[10] == 10*frame_nbytes
        assert len(ro) == 0

    @pytest.mark.parametrize('frame_nr,offset', [
        ([10], [5]),
        ([5, 15], [-1, 1])])
    def test_with_list(self, frame_nr, offset):
        ro = RawOffsets(frame_nr=frame_nr, offset=offset)
        assert len(ro) == len(frame_nr)
        assert ro[11] == offset[0]

    @pytest.mark.parametrize('invalid_frame_nbytes', [1.5, (4,)])
    def test_invalid_frame_nbytes(self, invalid_frame_nbytes):
        with pytest.raises(TypeError):
            RawOffsets(frame_nbytes=invalid_frame_nbytes)

    @pytest.mark.parametrize('frame_nr,offset', [
        ([1], None),
        ([5, 15], [0])])
    def test_invalid_lists(self, frame_nr, offset):
        with pytest.raises(ValueError):
            RawOffsets(frame_nr=frame_nr, offset=offset)

    @pytest.mark.parametrize('frame_nbytes', [0, 10])
    def test_set_offsets(self, frame_nbytes):
        ro = RawOffsets(frame_nbytes=frame_nbytes)
        assert ro[10] == 10*frame_nbytes
        ro[10] = 1 + 10*frame_nbytes
        assert len(ro) == 1
        assert ro.frame_nr == [10]
        assert ro.offset == [1]
        assert ro[10] == 1 + 10*frame_nbytes
        assert ro[11] == 1 + 11*frame_nbytes
        assert ro[9] == 0 + 9*frame_nbytes
        assert ro[15] == 1 + 15*frame_nbytes
        ro[14] = 10 + 14*frame_nbytes
        assert len(ro) == 2
        assert ro.frame_nr == [10, 14]
        assert ro.offset == [1, 10]
        assert ro[15] == 10 + 15*frame_nbytes
        # If we add no new information, ignore
        ro[12] = 1 + 12*frame_nbytes
        assert len(ro) == 2
        assert ro.frame_nr == [10, 14]
        assert ro.offset == [1, 10]
        # Though if it precedes a frame with the same offset, we move.
        ro[8] = 1 + 8*frame_nbytes
        assert ro.frame_nr == [8, 14]
        assert ro.offset == [1, 10]
        assert ro[9] == 1 + 9*frame_nbytes



# Licensed under the GPLv3 - see LICENSE
"""
Base definitions for payloads.

Defines a payload class PayloadBase that can be used to hold the words
corresponding to a frame payload, providing access to the values encoded
in it as a numpy array.
"""
import operator
from functools import reduce

import numpy as np


__all__ = ['PayloadBase']


class PayloadBase:
    """Container for decoding and encoding baseband payloads.

    Any subclass should define dictionaries ``_decoders`` and ``_encoders``,
    which hold functions that decode/encode the payload words to/from ndarray.
    These dictionaries are assumed to be indexed by ``bps``.

    Parameters
    ----------
    words : `~numpy.ndarray`
        Array containg LSB unsigned words (with the right size) that
        encode the payload.
    header : header instance
        If given, used to infer the sample shape, bps, and whether
        the data are complex.
    sample_shape : tuple
        Shape of the samples (e.g., (nchan,)).  Default: ().
    bps : int
        Bits per elementary sample, i.e., per channel and per real or
        imaginary component.  Default: 2.
    complex_data : bool
        Whether the data are complex.  Default: `False`.
    """
    # Possible fixed payload size in bytes.
    _nbytes = None
    # Default for whether to memmap payload
    _memmap = False
    # Default type for encoded data.
    _dtype_word = np.dtype('<u4')
    """Default for words: 32-bit unsigned integers, with lsb first."""
    # To be defined by subclasses.
    _encoders = {}
    _decoders = {}
    # Placeholder for sample shape named tuple.
    _sample_shape_maker = None

    def __init__(self, words, *, header=None,
                 sample_shape=(), bps=2, complex_data=False):
        if header is not None:
            sample_shape = header.sample_shape
            bps = header.bps
            complex_data = header.complex_data
            if self._nbytes is None:
                self._nbytes = header.payload_nbytes
            elif self._nbytes != header.payload_nbytes:
                raise ValueError("header payload size should be {0}"
                                 .format(self._nbytes))

        self.words = words
        if self._sample_shape_maker is not None:
            self.sample_shape = self._sample_shape_maker(*sample_shape)
        else:
            self.sample_shape = sample_shape
        self._sample_size = reduce(operator.mul, sample_shape, 1)
        self.bps = bps
        self.complex_data = complex_data
        self._bpfs = bps * (2 if complex_data else 1) * self._sample_size
        self._coder = bps
        if self._nbytes is not None and self._nbytes != words.nbytes:
            raise ValueError("encoded data should have length {0}"
                             .format(self._nbytes))
        if words.dtype != self._dtype_word:
            raise ValueError("encoded data should have dtype {0}"
                             .format(self._dtype_word))

    @classmethod
    def fromfile(cls, fh, header=None, *, payload_nbytes=None,
                 dtype=None, memmap=None, **kwargs):
        """Read payload from filehandle and decode it into data.

        Parameters
        ----------
        fh : filehandle
            From which data is read.
        header : header instance, optional
            If given, used to infer ``payload_nbytes``, ``bps``,
            ``sample_shape``, and ``complex_data``.  If not given, those have
            to be passed in.
        payload_nbytes : int, optional
            Number of bytes to read.  Except for fixed-length payloads,
            required if no ``header`` is given.
        dtype : `~numpy.dtype`, optional
            Type of words to read.  Default: taken from class attribute.
        memmap : bool, optional
            If `False`, read from file.  Otherwise, map the file in memory
            (see `~numpy.memmap`).  Only useful for large payloads.
            Default: taken from class attribute.

        Any other (keyword) arguments are passed on to the class initialiser.
        """
        if header is not None:
            payload_nbytes = header.payload_nbytes
            kwargs['header'] = header
        elif payload_nbytes is None:
            payload_nbytes = cls._nbytes
            if payload_nbytes is None:
                raise ValueError(
                    "payload_nbytes or header should be passed in "
                    "if no default payload size is defined on the class.")
        if dtype is None:
            dtype = cls._dtype_word
        if memmap is None:
            memmap = cls._memmap

        if memmap:
            shape = (payload_nbytes // dtype.itemsize,)
            if hasattr(fh, 'memmap'):
                words = fh.memmap(dtype=dtype, shape=shape)
            else:
                mode = fh.mode.replace('b', '')
                offset = fh.tell()
                words = np.memmap(fh, mode=mode, dtype=dtype,
                                  offset=offset, shape=shape)
                fh.seek(offset + words.nbytes)

        else:
            s = fh.read(payload_nbytes)
            if len(s) < payload_nbytes:
                raise EOFError("could not read full payload.")
            words = np.frombuffer(s, dtype=dtype)

        return cls(words, **kwargs)

    def tofile(self, fh):
        """Write payload to filehandle."""
        return fh.write(self.words.tobytes())

    @classmethod
    def fromdata(cls, data, header=None, bps=2):
        """Encode data as a payload.

        Parameters
        ----------
        data : `~numpy.ndarray`
            Data to be encoded, either complex or real. The trailing
            dimensions are used to infer ``sample_shape``.
        header : header instance, optional
            If given, used to infer to get ``bps``.
        bps : int, optional
            Bits per elementary sample, i.e., per channel and per real or
            imaginary component, used if header is not given.  Default: 2.
        """
        sample_shape = data.shape[1:]
        complex_data = data.dtype.kind == 'c'
        if header:
            bps = header.bps
            if header.sample_shape != sample_shape:
                raise ValueError(f"header is for sample_shape={sample_shape} "
                                 f"but data has {data.shape[1:]}")
            if header.complex_data != complex_data:
                raise ValueError("header is for {0} data but data are {1}"
                                 .format(*(('complex' if c else 'real') for c
                                           in (header.complex_data,
                                               complex_data))))
        try:
            encoder = cls._encoders[bps]
        except KeyError:
            raise ValueError("{0} cannot encode data with {1} bits"
                             .format(cls.__name__, bps))
        if complex_data:
            data = data.view((data.real.dtype, (2,)))
        words = encoder(data).ravel().view(cls._dtype_word)
        return cls(words, sample_shape=sample_shape, bps=bps,
                   complex_data=complex_data)

    def __array__(self, dtype=None):
        """Interface to arrays."""
        if dtype is None or dtype == self.dtype:
            return self.data
        else:
            return self.data.astype(dtype)

    @property
    def nbytes(self):
        """Size of the payload in bytes."""
        return self.words.nbytes

    def __len__(self):
        """Number of samples in the payload."""
        return self.words.nbytes * 8 // self._bpfs

    @property
    def shape(self):
        """Shape of the decoded data array."""
        return (len(self),) + self.sample_shape

    @property
    def size(self):
        """Total number of component samples in the decoded data array."""
        return len(self) * self._sample_size

    @property
    def ndim(self):
        """Number of dimensions of the decoded data array."""
        return 1 + len(self.sample_shape)

    @property
    def dtype(self):
        """Numeric type of the decoded data array."""
        return np.dtype(np.complex64 if self.complex_data else np.float32)

    def _item_to_slices(self, item):
        """Get word and data slices required to obtain given item.

        Parameters
        ----------
        item : int, slice, or tuple
            Sample indices.  An int represents a single sample, a slice
            a sample range, and a tuple of ints/slices a range for
            multi-channel data.

        Returns
        -------
        words_slice : slice
            Slice such that if one decodes ``ds = self.words[words_slice]``,
            ``ds`` is the smallest possible array that includes all
            of the requested ``item``.
        data_slice : int or slice
            Int or slice such that ``decode(ds)[data_slice]`` is the requested
            ``item``.

        Notes
        -----
        ``item`` is restricted to (tuples of) ints or slices, so one cannot
        access non-contiguous samples using advanced indexing.  If ``item``
        is a slice, a negative increment cannot be used.  The function is
        unable to parse payloads whose words have unused space (eg. VDIF files
        with 20 bits/sample).
        """
        if isinstance(item, tuple):
            sample_index = item[1:]
            item = item[0]
        else:
            sample_index = ()

        nsample = len(self)
        is_slice = isinstance(item, slice)
        if is_slice:
            start, stop, step = item.indices(nsample)
            assert step > 0, "cannot deal with negative steps yet."
            n = stop - start
            if step == 1:
                step = None
        else:
            try:
                item = operator.index(item)
            except Exception:
                raise TypeError("{0} object can only be indexed or sliced."
                                .format(type(self)))
            if item < 0:
                item += nsample

            if not (0 <= item < nsample):
                raise IndexError("{0} index out of range.".format(type(self)))

            start, stop, step, n = item, item+1, 1, 1

        if n == nsample:
            words_slice = slice(None)
            data_slice = slice(None, None, step) if is_slice else 0

        else:
            bpw = 8 * self.words.itemsize
            bpfs = self._bpfs
            if bpfs % bpw == 0:
                # Each full sample requires one or more encoded words.
                # Get corresponding range in words required, and decode those.
                wpfs = bpfs // bpw
                words_slice = slice(start * wpfs, stop * wpfs)
                data_slice = slice(None, None, step) if is_slice else 0

            elif bpw % bpfs == 0:
                # Each word contains multiple samples.
                # Get words in which required samples are contained.
                fspw = bpw // bpfs
                w_start, o_start = divmod(start, fspw)
                w_stop, o_stop = divmod(stop, fspw)

                words_slice = slice(w_start, w_stop + 1 if o_stop else w_stop)
                data_slice = slice(o_start if o_start else None,
                                   o_start + n if o_stop else None,
                                   step) if is_slice else o_start

            else:
                raise TypeError("do not know how to extract data when full "
                                "samples have {0} bits and words have {1} bits"
                                .format(bpfs, bpw))

        return words_slice, (data_slice,) + sample_index

    def __getitem__(self, item=()):
        decoder = self._decoders[self._coder]
        if item == () or item == slice(None):
            data = decoder(self.words)
            if self.complex_data:
                data = data.view(self.dtype)
            return data.reshape(self.shape)

        words_slice, data_slice = self._item_to_slices(item)

        return (decoder(self.words[words_slice]).view(self.dtype)
                .reshape(-1, *self.sample_shape)[data_slice])

    def __setitem__(self, item, data):
        if item == () or item == slice(None):
            words_slice = data_slice = slice(None)
        else:
            words_slice, data_slice = self._item_to_slices(item)

        data = np.asanyarray(data)
        # Check if the new data spans an entire word and is correctly shaped.
        # If so, skip decoding.  If not, decode appropriate words and insert
        # new data.
        if not (data_slice == slice(None)
                and data.shape[-len(self.sample_shape):] == self.sample_shape
                and data.dtype.kind == self.dtype.kind):
            decoder = self._decoders[self._coder]
            current_data = decoder(self.words[words_slice])
            if self.complex_data:
                current_data = current_data.view(self.dtype)
            current_data.shape = (-1,) + self.sample_shape
            current_data[data_slice] = data
            data = current_data

        if data.dtype.kind == 'c':
            data = data.view((data.real.dtype, (2,)))

        encoder = self._encoders[self._coder]
        self.words[words_slice] = encoder(data).ravel().view(self._dtype_word)

    data = property(__getitem__, doc="Full decoded payload.")

    def __eq__(self, other):
        return (type(self) is type(other)
                and self.shape == other.shape
                and self.dtype == other.dtype
                and (self.words is other.words
                     or np.all(self.words == other.words)))

    def __ne__(self, other):
        return not self.__eq__(other)



# Licensed under the GPLv3 - see LICENSE
"""GMRT Software Backend (GSB) data reader.

See http://gmrt.ncra.tifr.res.in/gmrt_hpage/sub_system/gmrt_gsb/index.htm
"""
from .base import open, info  # noqa
from .header import GSBHeader  # noqa
from .payload import GSBPayload  # noqa
from .frame import GSBFrame  # noqa
